#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "macros.h"
#include "k_means.h"

static void
printf_matrix(const double *m, int ncol, int nrow)
{
	const double *ptr = m;
	int i, j;
	printf("\n");
	for (i = 0; i < nrow; i++) {
		for (j = 0; j < ncol; j++) {
			printf("%f ", *ptr);
			ptr++;
		}
		printf("\n");
	}
}

static double
euclid_distance(const double *x1, const double *x2, int dim)
{
	const double *p1, *p2;
	double tmp, d = 0.0;
	int i;
	p1 = x1;
	p2 = x2;

	for (i = 0; i < dim; i++) {
		tmp = *p1 - *p2;
		d += tmp * tmp;
		p1++;
		p2++;
	}

	return sqrt(d);
}
struct k_means_context*
k_means_new()
{
	struct k_means_context* ptr;
	ptr = malloc(sizeof(struct k_means_context));
	return_val_if_fail(ptr != NULL, ptr);
	return ptr;
}

void k_means_init_(struct k_means_context *ctx)
{
	return_if_fail(ctx != NULL);

	int dim, size, k;
	k    = ctx->k;
	dim  = ctx->dim;
	size = k * dim;
	ctx->means = malloc((2 * size + k) * sizeof(double));
	return_if_fail(ctx->means != NULL);

	memset(ctx->means, 0, (2 * size + k) * sizeof(double));
	ctx->dists = ctx->means + size;
	ctx->weight = ctx->dists + size;
}

void k_means_free(struct k_means_context **ctx)
{
	return_if_fail(ctx != NULL);
	return_if_fail(*ctx != NULL);
	free((*ctx)->means);
	free(*ctx);
	*ctx = NULL;
}

void k_means_clear(struct k_means_context *ctx)
{
	free(ctx->means);
	ctx->k = 0;
	ctx->dim = 0;
}

/* Set number of mean. */
void
k_means_set_nmean(struct k_means_context *ctx, int nmean)
{
	ctx->k = nmean;
}
/* Get number of mean. */
int
k_means_get_nmean(struct k_means_context *ctx)
{
	return ctx->k;
}

/* Set fitted data item dimension. */
void
k_means_set_dimension(struct k_means_context *ctx, int dim)
{
	ctx->dim = dim;
}
/* Get fitted data item dimension. */
int
k_means_get_dimension(struct k_means_context *ctx)
{
	return ctx->dim;
}

/* Set n-th mean component weight. */
void
k_means_set_nth_weight(struct k_means_context *ctx, int nth, double weight)
{
	ctx->weight[nth] = weight;
}
/* Get n-th mean component weight. */
double
k_means_get_nth_weight(struct k_means_context *ctx, int nth)
{
	return ctx->weight[nth];
}

/* Set n-th mean. */
void
k_means_set_nth_mean(struct k_means_context *ctx, int nth, const double *mu)
{
	memcpy(ctx->means + nth * ctx->dim, mu, ctx->k * sizeof(double));
}
/* Get n-th mean. */
void
k_means_get_nth_mean(const struct k_means_context *ctx, int nth, double *mu)
{
	memcpy(mu, ctx->means + nth * ctx->dim, ctx->k * sizeof(double));
}

/* Set n-th distance array. */
void
k_means_set_nth_distance(struct k_means_context *ctx, int nth, const double *dist)
{
	memcpy(ctx->means + nth * ctx->dim, dist, ctx->k * sizeof(double));
}
/* Get n-th distance array. */
void
k_means_get_nth_distance(const struct k_means_context *ctx, int nth, double *dist)
{
	memcpy(dist, ctx->means + nth * ctx->dim, ctx->k * sizeof(double));
}

int
k_means(double *means, int nmeans, double *x, int nx, int dim, double epsilon)
{
	double dist, min_dist, *data, *mean, *tmp_mean, *ptr;
	int i, j, k, m_len, done = 1, ncluster, *clusters, *nclusters;

	k   = nmeans;

	m_len = k * dim * sizeof(double);
	clusters = malloc(k * sizeof(int));
	nclusters = malloc(nx * sizeof(int));
	tmp_mean = malloc(m_len);

	while (done) {
		data = x;

		memset(tmp_mean, 0, m_len);
		memset(clusters, 0, k * sizeof(int));
		for (i = 0; i < nx; i++) {
			mean = means;
			min_dist = euclid_distance(data, mean, dim);
			ncluster = 0;
			for (j = 1; j < k; j++) {
				mean += dim;
				dist = euclid_distance(data, mean, dim);
				if (dist < min_dist) {
					min_dist = dist;
					ncluster = j;
				}
			}
			ptr = tmp_mean + ncluster * dim;
			clusters[ncluster]++;
			nclusters[i] = ncluster;
			for (j = 0; j < dim; j++) {
				*ptr += *data;
				ptr++;
				data++;
			}
		}
		ptr = tmp_mean;
		mean = means;

		dist = 0.0;
		for (i = 0; i < k; i++) {
			for (j = 0; j < dim; j++) {
				if (clusters[i] > 0)
					*ptr /= clusters[i];
				else
					*ptr = *mean;
				dist += fabs(*ptr - *mean);
				ptr++;
				mean++;
			}
		}

		done = dist > 0.001;
		memcpy(means, tmp_mean, m_len);
	}

	mean = means;
	for (i = 0; i < k; i++) {
		printf_matrix(mean, dim, 1);
		mean += dim;
	}

	free(clusters);
	free(tmp_mean);
	free(nclusters);
	return 0;
}

/*
 * Converge the k-means-algorithm, repeating k-means-steps until.
 * Data in array mast be dispose of row range [x, y, z, x, y, z,...]
 */
int
k_means_converge(struct k_means_context *ctx, double *x, int nx, double epsilon)
{
	double dist, min_dist, *data, *mean, *tmp_mean, *ptr;
	int i, j, k, m_len, dim, done = 1, ncluster, *clusters, *nclusters;

	k   = ctx->k;
	dim = ctx->dim;
	m_len = k * dim * sizeof(double);
	clusters = malloc(k * sizeof(int));
	nclusters = malloc(nx * sizeof(int));
	tmp_mean = malloc(m_len);
	memset(ctx->dists, 0, m_len);

	while (done) {
		data = x;

		memset(tmp_mean, 0, m_len);
		memset(clusters, 0, k * sizeof(int));
		for (i = 0; i < nx; i++) {
			mean = ctx->means;
			min_dist = euclid_distance(data, mean, dim);
			ncluster = 0;
			for (j = 1; j < k; j++) {
				mean += dim;
				dist = euclid_distance(data, mean, dim);
				if (dist < min_dist) {
					min_dist = dist;
					ncluster = j;
				}
			}
			ptr = tmp_mean + ncluster * dim;
			clusters[ncluster]++;
			nclusters[i] = ncluster;
			for (j = 0; j < dim; j++) {
				*ptr += *data;
				ptr++;
				data++;
			}
		}
		ptr = tmp_mean;
		mean = ctx->means;

		dist = 0.0;
		for (i = 0; i < k; i++) {
			for (j = 0; j < dim; j++) {
				if (clusters[i] > 0)
					*ptr /= clusters[i];
				else
					*ptr = *mean;
				dist += fabs(*ptr - *mean);
				ptr++;
				mean++;
			}
		}

		done = dist > 0.001;
		memcpy(ctx->means, tmp_mean, m_len);
	}
	ptr = ctx->weight;
	for (i = 0; i < k; i++) {
		*ptr = (double)clusters[i] / nx;
		ptr++;
	}

	data = x;
	for (i = 0; i < nx; i++) {
		ncluster = nclusters[i];
		mean = ctx->means + ncluster * dim;
		ptr = ctx->dists + ncluster * dim;
		for (j = 0; j < dim; j++) {
			dist = fabs(*mean - *data);
			if (dist > *ptr)
				*ptr = dist;
			mean++;
			data++;
			ptr++;
		}
	}
ptr = ctx->dists;
mean = ctx->means;
for (i = 0; i < k; i++) {
	printf_matrix(ptr, dim, 1);
	printf_matrix(mean, dim, 1);
	ptr += dim;
	mean += dim;
}
	free(clusters);
	free(tmp_mean);
	free(nclusters);
	return 0;
}

/*
 * Determine if a sample x represents model or not.
 * return: k_means_SUCCESS - if x represents model and data is valid
 *         k_means_ERROR - else
 */
int
k_means_sample_to_match_model(struct k_means_context *ctx, double *x)
{
	double d, *m, *dist, *ptr;
	int i, j, k, dim, res;

	dist = ctx->dists;
	m    = ctx->means;
	k    = ctx->k;
	dim  = ctx->dim;

	for (i = 0; i < k; i++) {
		ptr = x;
		res = 1;
		for (j = 0; j < dim; j++) {
			d = fabs(*ptr - *m);
			res *= (d < *dist + 10);
			m++;
			dist++;
			ptr++;
		}
		if (res)
			return k_means_SUCCESS;
	}
	return k_means_ERROR;
}

/*
 * Update k_means_context using new sample data
 */
void
k_means_update(struct k_means_context *ctx, double *x)
{
	double d, *m1, *m2, *dist, *ptr, a1, a2, tmp;
	int i, j, k, dim;

	dist = ctx->dists;
	m2 = m1   = ctx->means;
	k    = ctx->k;
	dim  = ctx->dim;

	for (i = 0; i < k; i++) {
		ptr = x;
		d = 0.0;
		for (j = 0; j < dim; j++) {
			tmp = fabs(*ptr - *m1);
			d += tmp * tmp / (*dist + 1);
			m1++;
			dist++;
			ptr++;
		}

		a1 = exp(-d);
		a2 = 1 - a1;
		ptr = x;
		for (j = 0; j < dim; j++) {
			*m2 = a2 * (*m2) + a1 * (*ptr);
			m2++;
			ptr++;
			
		}
	}
}

void
k_means_reset(struct k_means_context *ctx, double *data, int data_size)
{
	double *w, *m, *d,  weight;
	int i, k, dim, r;
	k        = ctx->k;
	dim      = ctx->dim;

	w = ctx->weight;
	m = ctx->means;
	d = ctx->dists;

	weight = 1.0 / k;

	for (i = 0; i < k; i++) {
		*w = weight;
		memset(d, 0, dim * sizeof(double));
		if (data != NULL && data_size > 0) {
			r = rand() % data_size;
			memcpy(m, data + r, dim * sizeof(double));
		} else {
			memset(m, 0, dim * sizeof(double));
		}
		w++;
		m += dim;
		d += dim;
	}
}
